"""CP2K adapted interpreter and modder for custodian."""

import os

from pymatgen.io.cp2k.inputs import Cp2kInput

from custodian.ansible.actions import DictActions, FileActions
from custodian.ansible.interpreter import Modder
from custodian.cp2k.utils import cleanup_input

__author__ = "Nicholas Winner"
__version__ = "1.0"
__email__ = "nwinner@berkeley.edu"
__date__ = "October 2021"


class Cp2kModder(Modder):
    """
    Cp2kModder is a lightweight class for applying modifications to cp2k input files. It
    also supports modifications that are file operations (e.g. copying).
    """

    def __init__(self, filename="cp2k.inp", actions=None, strict=True, ci=None, directory="./") -> None:
        """Initialize a Modder for Cp2kInput sets.

        Args:
            filename (str): name of cp2k input file to modify. This file will be overwritten
                if actions are applied.
            actions ([Action]): A sequence of supported actions. See
                :mod:`custodian.ansible.actions`. Default is None,
                which means DictActions and FileActions are supported.
            strict (bool): Indicating whether to use strict mode. In non-strict
                mode, unsupported actions are simply ignored without any
                errors raised. In strict mode, if an unsupported action is
                supplied, a ValueError is raised. Defaults to True.
            ci (Cp2kInput): A Cp2kInput object from the current directory.
                Initialized automatically if not passed (but passing it will
                avoid having to re-parse the directory).
            directory (str): The directory containing the Cp2kInput set. Defaults to "./".
        """
        self.directory = directory
        self.ci = ci or Cp2kInput.from_file(os.path.join(self.directory, filename))
        self.filename = filename
        actions = actions or [FileActions, DictActions]
        super().__init__(actions, strict, directory=directory)

    def apply_actions(self, actions) -> None:
        """
        Applies a list of actions to the CP2K Input Set and rewrites modified
        files.

        Args:
            actions (dict): A list of actions of the form {'file': filename,
                'action': moddermodification} or {'dict': cp2k_key,
                'action': moddermodification}.
        """
        modified = []
        for action in actions:
            if "dict" in action:
                k = action["dict"]
                modified.append(k)
                Cp2kModder._modify(action["action"], self.ci)
            elif "file" in action:
                self.modify(action["action"], action["file"])
                self.ci = Cp2kInput.from_file(os.path.join(self.directory, self.filename))
            else:
                raise ValueError(f"Unrecognized format: {action}")
        cleanup_input(self.ci)
        self.ci.write_file(os.path.join(self.directory, self.filename))

    @staticmethod
    def _modify(modification, obj) -> None:
        """
        Note that modify makes actual in-place modifications. It does not
        return a copy.

        Args:
            modification (dict): Modification must be {action_keyword :
                settings}. E.g., {'_set': {'Hello':'Universe', 'Bye': 'World'}}
            obj (dict/str/object): Object to modify depending on actions. For
                example, for DictActions, obj will be a dict to be modified.
                For FileActions, obj will be a string with a full pathname to a
                file.
        """
        modification = list(modification.items()) if isinstance(modification, dict) else modification
        for action, settings in modification:
            try:
                getattr(obj, action[1:])(settings)
            except KeyError:
                continue



"""Implements various interpreters and modders for VASP."""

import os

from pymatgen.io.vasp.inputs import VaspInput

from custodian.ansible.actions import DictActions, FileActions
from custodian.ansible.interpreter import Modder


class VaspModder(Modder):
    """A Modder for VaspInputSets."""

    def __init__(self, actions=None, strict=True, vi=None, directory="./") -> None:
        """Initialize a Modder for VaspInput sets.

        Args:
            actions ([Action]): A sequence of supported actions. See
                :mod:`custodian.ansible.actions`. Default is None,
                which means DictActions and FileActions are supported.
            strict (bool): Indicating whether to use strict mode. In non-strict
                mode, unsupported actions are simply ignored without any
                errors raised. In strict mode, if an unsupported action is
                supplied, a ValueError is raised. Defaults to True.
            vi (VaspInput): A VaspInput object from the current directory.
                Initialized automatically if not passed (but passing it will
                avoid having to re-parse the directory).
            directory (str): The directory containing the VaspInput set.
        """
        self.vi = vi or VaspInput.from_directory(directory)
        self.directory = directory
        actions = actions or [FileActions, DictActions]
        super().__init__(actions, strict, directory=directory)

    def apply_actions(self, actions) -> None:
        """
        Applies a list of actions to the Vasp Input Set and rewrites modified
        files.

        Args:
            actions (dict): A list of actions of the form {'file': filename,
                'action': moddermodification} or {'dict': vaspinput_key,
                'action': moddermodification}.
        """
        modified = []
        for action in actions:
            if "dict" in action:
                k = action["dict"]
                modified.append(k)
                self.vi[k] = self.modify_object(action["action"], self.vi[k])
            elif "file" in action:
                self.modify(action["action"], action["file"])
            else:
                raise ValueError(f"Unrecognized format: {action}")
        for file in modified:
            self.vi[file].write_file(os.path.join(self.directory, file))



"""
This module defines various classes of supported actions. All actions are
implemented as static methods, but are defined using classes (as opposed to
modules) so that a set of well-defined actions can be namespaced easily.
"""

import os
import shutil


def get_nested_dict(input_dict, key):
    """Helper function to interpret a nested dict input."""
    current = input_dict
    tokens = key.split("->")
    n = len(tokens)
    for i, tok in enumerate(tokens):
        if tok not in current and i < n - 1:
            current[tok] = {}
        elif i == n - 1:
            return current, tokens[-1]
        current = current[tok]
    return None


class DictActions:
    """Class to implement the supported mongo-like modifications on a dict.

    Supported keywords include the following Mongo-based keywords, with the
    usual meanings (refer to Mongo documentation for information):

        _inc
        _set
        _unset
        _push
        _push_all
        _add_to_set (but _each is not supported)
        _pop
        _pull
        _pull_all
        _rename

    However, note that "_set" does not support modification of nested dicts
    using the mongo {"a.b":1} notation. This is because mongo does not allow
    keys with "." to be inserted. Instead, nested dict modification is
    supported using a special "->" keyword, e.g. {"a->b": 1}
    """

    @staticmethod
    def set(input_dict, settings, directory=None) -> None:
        """
        Sets a value using MongoDB syntax.

        Args:
            input_dict (dict): The input dictionary to be modified.
            settings (dict): The specification of the modification to be made.
            directory (None): dummy parameter for compatibility with FileActions
        """
        for key, val in settings.items():
            dct, sub_key = get_nested_dict(input_dict, key)
            dct[sub_key] = val

    @staticmethod
    def unset(input_dict, settings, directory=None) -> None:
        """
        Unset a value using MongoDB syntax.

        Args:
            input_dict (dict): The input dictionary to be modified.
            settings (dict): The specification of the modification to be made.
            directory (None): dummy parameter for compatibility with FileActions
        """
        for key in settings:
            dct, inner_key = get_nested_dict(input_dict, key)
            del dct[inner_key]

    @staticmethod
    def push(input_dict, settings, directory=None) -> None:
        """
        Push to a list using MongoDB syntax.

        Args:
            input_dict (dict): The input dictionary to be modified.
            settings (dict): The specification of the modification to be made.
            directory (None): dummy parameter for compatibility with FileActions
        """
        for key, val in settings.items():
            dct, sub_key = get_nested_dict(input_dict, key)
            if sub_key in dct:
                dct[sub_key].append(val)
            else:
                dct[sub_key] = [val]

    @staticmethod
    def push_all(input_dict, settings, directory=None) -> None:
        """
        Push multiple items to a list using MongoDB syntax.

        Args:
            input_dict (dict): The input dictionary to be modified.
            settings (dict): The specification of the modification to be made.
            directory (None): dummy parameter for compatibility with FileActions
        """
        for k1, val in settings.items():
            dct, k2 = get_nested_dict(input_dict, k1)
            if k2 in dct:
                dct[k2] += val
            else:
                dct[k2] = val

    @staticmethod
    def inc(input_dict, settings, directory=None) -> None:
        """
        Increment a value using MongoDB syntax.

        Args:
            input_dict (dict): The input dictionary to be modified.
            settings (dict): The specification of the modification to be made.
            directory (None): dummy parameter for compatibility with FileActions
        """
        for key, val in settings.items():
            dct, sub_key = get_nested_dict(input_dict, key)
            if sub_key in dct:
                dct[sub_key] += val
            else:
                dct[sub_key] = val

    @staticmethod
    def rename(input_dict, settings, directory=None) -> None:
        """
        Rename a key using MongoDB syntax.

        Args:
            input_dict (dict): The input dictionary to be modified.
            settings (dict): The specification of the modification to be made.
            directory (None): dummy parameter for compatibility with FileActions
        """
        for key, val in settings.items():
            if input_val := input_dict.pop(key, None):
                input_dict[val] = input_val

    @staticmethod
    def add_to_set(input_dict, settings, directory=None) -> None:
        """
        Add to set using MongoDB syntax.

        Args:
            input_dict (dict): The input dictionary to be modified.
            settings (dict): The specification of the modification to be made.
            directory (None): dummy parameter for compatibility with FileActions
        """
        for key, val in settings.items():
            dct, sub_key = get_nested_dict(input_dict, key)
            if sub_key in dct and (not isinstance(dct[sub_key], list)):
                raise ValueError(f"Keyword {key} does not refer to an array.")
            if sub_key in dct and val not in dct[sub_key]:
                dct[sub_key].append(val)
            elif sub_key not in dct:
                dct[sub_key] = val

    @staticmethod
    def pull(input_dict, settings, directory=None) -> None:
        """
        Pull an item using MongoDB syntax.

        Args:
            input_dict (dict): The input dictionary to be modified.
            settings (dict): The specification of the modification to be made.
            directory (None): dummy parameter for compatibility with FileActions
        """
        for k1, val in settings.items():
            dct, k2 = get_nested_dict(input_dict, k1)
            if k2 in dct and (not isinstance(dct[k2], list)):
                raise ValueError(f"Keyword {k1} does not refer to an array.")
            if k2 in dct:
                dct[k2] = [itm for itm in dct[k2] if itm != val]

    @staticmethod
    def pull_all(input_dict, settings, directory=None) -> None:
        """
        Pull multiple items to a list using MongoDB syntax.

        Args:
            input_dict (dict): The input dictionary to be modified.
            settings (dict): The specification of the modification to be made.
            directory (None): dummy parameter for compatibility with FileActions
        """
        for key, val in settings.items():
            if key in input_dict and (not isinstance(input_dict[key], list)):
                raise ValueError(f"Keyword {key} does not refer to an array.")
            for itm in val:
                DictActions.pull(input_dict, {key: itm})

    @staticmethod
    def pop(input_dict, settings, directory=None) -> None:
        """
        Pop item from a list using MongoDB syntax.

        Args:
            input_dict (dict): The input dictionary to be modified.
            settings (dict): The specification of the modification to be made.
            directory (None): dummy parameter for compatibility with FileActions
        """
        for key, val in settings.items():
            dct, sub_key = get_nested_dict(input_dict, key)
            if sub_key in dct and (not isinstance(dct[sub_key], list)):
                raise ValueError(f"Keyword {key} does not refer to an array.")
            if val == 1:
                dct[sub_key].pop()
            elif val == -1:
                dct[sub_key].pop(0)


class FileActions:
    """
    Class of supported file actions. For FileActions, the modder class takes in
    a filename as a string. The filename should preferably be a full path to
    avoid ambiguity.
    """

    @staticmethod
    def file_create(filename, settings, directory) -> None:
        """
        Creates a file.

        Args:
            filename (str): Filename.
            settings (dict): Must be {"content": actual_content}
            directory (str): Directory to create file in
        """
        if len(settings) != 1:
            raise ValueError("Settings must only contain one item with key 'content'.")
        for key, val in settings.items():
            if key == "content":
                with open(filename, "w") as file:
                    file.write(val)

    @staticmethod
    def file_move(filename, settings, directory) -> None:
        """
        Moves a file. {'_file_move': {'dest': 'new_file_name'}}.

        Args:
            filename (str): Filename.
            settings (dict): Must be {"dest": path of new file}
            directory (str): Directory to move file from and to
        """
        if len(settings) != 1:
            raise ValueError("Settings must only contain one item with key 'dest'.")
        for key, val in settings.items():
            if key == "dest":
                shutil.move(os.path.join(directory, filename), os.path.join(directory, val))

    @staticmethod
    def file_delete(filename, settings, directory) -> None:
        """
        Deletes a file. {'_file_delete': {'mode': "actual"}}.

        Args:
            filename (str): Filename.
            settings (dict): Must be {"mode": actual/simulated}. Simulated
                mode only prints the action without performing it.
            directory (str): Directory to delete file in
        """
        if len(settings) != 1:
            raise ValueError("Settings must only contain one item with key 'mode'.")
        for key, val in settings.items():
            if key == "mode" and val == "actual":
                try:
                    os.remove(os.path.join(directory, filename))
                except OSError:
                    # Skip file not found error.
                    pass
            elif key == "mode" and val == "simulated":
                print(f"Simulated removal of {filename}")

    @staticmethod
    def file_copy(filename, settings, directory) -> None:
        """
        Copies a file. {'_file_copy': {'dest': 'new_file_name'}}.

        Args:
            filename (str): Filename.
            settings (dict): Must be {"dest": path of new file}
            directory (str): Directory to copy file to/from
        """
        for key, val in settings.items():
            if key.startswith("dest"):
                shutil.copyfile(os.path.join(directory, filename), os.path.join(directory, val))

    @staticmethod
    def file_modify(filename, settings, directory) -> None:
        """
        Modifies file access.

        Args:
            filename (str): Filename.
            settings (dict): Can be "mode" or "owners"
            directory (str): Directory to modify file in
        """
        for key, val in settings.items():
            if key == "mode":
                os.chmod(os.path.join(directory, filename), val)
            if key == "owners":
                # TODO fix this mypy error, missing 3rd positional argument to chown
                os.chown(os.path.join(directory, filename), val)  # type: ignore[call-arg]



"""This module implements basic kinds of jobs for Gaussian runs."""

from __future__ import annotations

import logging
import os
import shutil
import subprocess
from fnmatch import filter
from typing import TYPE_CHECKING

from pymatgen.io.gaussian import GaussianInput, GaussianOutput

from custodian.custodian import Job
from custodian.gaussian.handlers import GaussianErrorHandler

if TYPE_CHECKING:
    from collections.abc import Generator

__author__ = "Rasha Atwi"
__version__ = "0.1"
__maintainer__ = "Rasha Atwi"
__email__ = "rasha.atwi@stonybrook.edu"
__status__ = "Alpha"
__date__ = "5/13/21"

logger = logging.getLogger(__name__)


class GaussianJob(Job):
    """A basic Gaussian job."""

    def __init__(
        self,
        gaussian_cmd: str,
        input_file: str,
        output_file: str,
        stderr_file: str = "stderr.txt",
        suffix: str = "",
        backup: bool = True,
    ):
        """
        Args:
            gaussian_cmd (str): Command to run Gaussian.
            input_file (str): Name of the Gaussian input file.
            output_file (str): Name of the Gaussian output file.
            stderr_file (str): Name of the stderr file. Defaults to 'stderr.txt'.
            suffix (str): String to append to the input and output files, e.g., to
                rename Gaussian input and output files from say gaussian.com and
                gaussian.out, to gaussian.com.guess1 and gaussian.out.guess1,
                respectively, provide ".guess1" as the suffix. . Defaults to ''.
            backup (bool): Whether to backup the input file. If True, the input will be
                copied with a ".orig" appended to the name. Defaults to True.
        """
        self.gaussian_cmd = gaussian_cmd
        self.input_file = input_file
        self.output_file = output_file
        self.stderr_file = stderr_file
        self.suffix = suffix
        self.backup = backup
        self.process: subprocess.Popen | None = None

    def setup(self, directory: str = "./") -> None:
        """
        Perform initial setup for the job, i.e., make a backup of the input file if
        requested.

        Args:
            directory (str): Directory where the job will be run. Defaults to './'.
        """
        if self.backup:
            shutil.copy(
                os.path.join(directory, self.input_file),
                os.path.join(directory, f"{self.input_file}.orig"),
            )

    def run(self, directory: str = "./") -> subprocess.Popen:
        """
        Perform the actual Gaussian run.

        Args:
            directory (str): Directory where the job will be run. Defaults to './'.

        Returns:
            subprocess.Popen:
                The process running the Gaussian job.
        """
        logger.info(f"Running command: {self.gaussian_cmd}")
        with (
            open(os.path.join(directory, self.output_file), "w") as out_file,
            open(os.path.join(directory, self.stderr_file), "w", buffering=1) as error_file,
        ):
            process = subprocess.Popen(self.gaussian_cmd, stdout=out_file, stderr=error_file, shell=True)
        self.process = process
        return process

    def postprocess(self, directory: str = "./") -> None:
        """
        Perform any postprocessing of the Gaussian run. This includes making a copy
        of the input and output file if a suffix is specified.

        Args:
            directory (str): Directory where the job was run. Defaults to './'.
        """
        for file in [self.input_file, self.output_file]:
            file_path = os.path.join(directory, file)
            if os.path.exists(file_path) and self.suffix != "":
                shutil.copy(file_path, f"{file_path}{self.suffix}")

    def terminate(self, directory: str = "./") -> None:
        """
        Terminate the Gaussian job.

        Args:
            directory (str): Directory where the job was run. Defaults to './'.
        """
        if self.process:
            self.process.terminate()

    @classmethod
    def generate_better_guess(
        cls,
        gaussian_cmd: str,
        input_file: str,
        output_file: str,
        stderr_file: str = "stderr.txt",
        backup: bool = True,
        cart_coords: bool = True,
        directory: str = "./",
    ) -> Generator[GaussianJob, None, None]:
        """
        Generate a better initial guess for a Gaussian calculation (optimization or
        SCF run). This is done by running the job at a lower level of theory
        (e.g., STO-3G). The geometry for this job is read from the checkpoint file
        generated by the previous job.

        Args:
            gaussian_cmd (str): Command to run Gaussian.
            input_file (str): Name of the Gaussian input file.
            output_file (str): Name of the Gaussian output file.
            stderr_file (str): Name of the stderr file. Defaults to 'stderr.txt'.
            backup (bool): Whether to backup the input file. If True, the input will be
                copied with a ".orig" appended to the name. Defaults to True.
            cart_coords (bool): Whether to use Cartesian coordinates in the input file.
                Defaults to True.
            directory (str): Directory where the job will be run. Defaults to './'.

        Yields:
            GaussianJob: The Gaussian job instance.
        """
        orig_input = GaussianInput.from_file(os.path.join(directory, input_file))
        yield (
            GaussianJob(
                gaussian_cmd=gaussian_cmd,
                input_file=input_file,
                output_file=output_file,
                stderr_file=stderr_file,
                suffix=".guess1",
                backup=backup,
            )
        )
        if GaussianErrorHandler.activate_better_guess:
            # TODO: check why it comes here only if the lower job is not
            #  failing and not in the else condition
            # continue only if other corrections are invalid or failed
            lower_output = GaussianOutput(os.path.join(directory, output_file))
            if len(lower_output.errors) == 0:
                # if the calculation at the lower level of theory succeeded
                if not filter(os.listdir("."), "*.[Cc][Hh][Kk]"):
                    raise FileNotFoundError("Missing checkpoint file. Required to read initial guesses")

                gin = GaussianInput(
                    mol=None,
                    charge=orig_input.charge,
                    spin_multiplicity=orig_input.spin_multiplicity,
                    title=orig_input.title,
                    functional=orig_input.functional,
                    basis_set=orig_input.basis_set,
                    route_parameters=lower_output.route_parameters,
                    input_parameters=orig_input.input_parameters,
                    link0_parameters=orig_input.link0_parameters,
                    dieze_tag=orig_input.dieze_tag,
                    gen_basis=orig_input.gen_basis,
                )
                gin.route_parameters["Guess"] = "Read"
                gin.route_parameters["Geom"] = "Checkpoint"
                gin.write_file(os.path.join(directory, input_file), cart_coords=cart_coords)

                yield (
                    GaussianJob(
                        gaussian_cmd=gaussian_cmd,
                        input_file=input_file,
                        output_file=output_file,
                        stderr_file=stderr_file,
                        suffix=".guess2",
                        backup=backup,
                    )
                )
            else:
                logger.info("Failed to generate a better initial guess")
        else:
            logger.info("Calculation completed normally without having to generate a better initial guess")



"""This module implements basic kinds of jobs for QChem runs."""

from __future__ import annotations

import copy
import os
import shutil
import subprocess
import warnings
from pathlib import Path

import numpy as np
from pymatgen.core import Molecule
from pymatgen.io.qchem.inputs import QCInput
from pymatgen.io.qchem.outputs import QCOutput, check_for_structure_changes
from pymatgen.io.qchem.sets import OptSet

from custodian.custodian import Job
from custodian.qchem.utils import perturb_coordinates, vector_list_diff

try:
    from openbabel import openbabel as ob
except ImportError:
    ob = None

__author__ = "Samuel Blau, Brandon Wood, Shyam Dwaraknath, Evan Spotte-Smith"
__copyright__ = "Copyright 2018, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Samuel Blau"
__email__ = "samblau1@gmail.com"
__status__ = "Alpha"
__date__ = "3/20/18"
__credits__ = "Xiaohui Qu"


class QCJob(Job):
    """A basic QChem Job."""

    def __init__(
        self,
        qchem_command,
        max_cores,
        multimode="openmp",
        input_file="mol.qin",
        output_file="mol.qout",
        qclog_file="mol.qclog",
        suffix="",
        calc_loc=None,
        nboexe=None,
        save_scratch=False,
        backup=True,
    ) -> None:
        """
        Args:
            qchem_command (str): Command to run QChem.
            max_cores (int): Maximum number of cores to parallelize over.
            multimode (str): Parallelization scheme, either openmp or mpi.
            input_file (str): Name of the QChem input file.
            output_file (str): Name of the QChem output file.
            qclog_file (str): Name of the file to redirect the standard output
                to. None means not to record the standard output.
            suffix (str): String to append to the file in postprocess.
            calc_loc (str): Path where Q-Chem should run. Defaults to None, in
                which case Q-Chem will run in the system-defined QCLOCALSCR.
            nboexe (str): Path to the NBO7 executable. Defaults to None.
            save_scratch (bool): Whether to save full scratch directory contents.
                Defaults to False.
            backup (bool): Whether to backup the initial input file. If True, the
                input will be copied with a ".orig" appended. Defaults to True.
        """
        if ob is None:
            raise RuntimeError("ERROR: Openbabel must be installed in order to use Q-Chem Custodian!")
        try:
            self.qchem_command = qchem_command.split(" ")
        except AttributeError:
            if isinstance(qchem_command, list):
                for val in qchem_command:
                    if not isinstance(val, str):
                        raise ValueError("Must either pass a string or a list of strings")
            else:
                raise ValueError("Must either pass a string or a list of strings")
            self.qchem_command = qchem_command
        self.multimode = multimode
        self.input_file = input_file
        self.output_file = output_file
        self.max_cores = max_cores
        self.qclog_file = qclog_file
        self.suffix = suffix
        self.calc_loc = calc_loc
        self.nboexe = nboexe
        self.save_scratch = save_scratch
        self.backup = backup

        try:
            slurm_cores = int(os.environ["SLURM_CPUS_ON_NODE"])
            if slurm_cores < self.max_cores:
                self.max_cores = slurm_cores
                print("max_cores reduced from", max_cores, "to", self.max_cores)
            else:
                print("max_cores remain at", self.max_cores)
        except KeyError:
            print("SLURM_CPUS_ON_NODE not in environment")

    @property
    def current_command(self, directory: str | Path = "./"):
        """The command to run QChem.

        Args:
            directory (str): The directory to run in. Defaults to "./".

        Returns:
            (str) The command to run QChem.
        """
        self._input_path = os.path.join(directory, self.input_file)
        self._output_path = os.path.join(directory, self.output_file)
        multi = {"openmp": "-nt", "mpi": "-np"}
        if self.multimode not in multi:
            raise RuntimeError("ERROR: Multimode should only be set to openmp or mpi")
        command = [multi[self.multimode], str(self.max_cores), self._input_path, self._output_path, "scratch"]
        command = self.qchem_command + command
        return " ".join(command)

    def setup(self, directory: str | Path = "./") -> None:
        """Sets up environment variables necessary to efficiently run QChem.

        Args:
            directory (str): The directory to run in. Defaults to "./".
        """
        self._input_path = os.path.join(directory, self.input_file)
        if self.backup:
            shutil.copy(self._input_path, os.path.join(directory, f"{self.input_file}.orig"))
        if self.multimode == "openmp":
            os.environ["QCTHREADS"] = str(self.max_cores)
            os.environ["OMP_NUM_THREADS"] = str(self.max_cores)
        os.environ["QCSCRATCH"] = str(Path(directory).resolve())
        if self.calc_loc is not None:
            os.environ["QCLOCALSCR"] = self.calc_loc
        qcinp = QCInput.from_file(self._input_path)
        if (
            qcinp.rem.get("run_nbo6", "none").lower() == "true"
            or qcinp.rem.get("nbo_external", "none").lower() == "true"
        ):
            os.environ["KMP_INIT_AT_FORK"] = "FALSE"
            if self.nboexe is None:
                raise RuntimeError("Trying to run NBO7 without providing NBOEXE in fworker! Exiting...")
            os.environ["NBOEXE"] = self.nboexe

    def postprocess(self, directory: str | Path = "./") -> None:
        """Renames and removes scratch files after running QChem.

        Args:
            directory (str): The directory to run in. Defaults to "./".
        """
        self._input_path = os.path.join(directory, self.input_file)
        self._output_path = os.path.join(directory, self.output_file)
        self._qclog_path = os.path.join(directory, self.qclog_file)
        scratch_dir = os.path.join(os.environ["QCSCRATCH"], "scratch")
        for file in ("HESS", "GRAD", "plots/dens.0.cube", "131.0", "53.0", "132.0"):
            file_path = os.path.join(scratch_dir, file)
            if os.path.isfile(file_path):
                shutil.copy(file_path, directory)
        if self.suffix != "":
            shutil.move(self._input_path, os.path.join(directory, self.input_file + self.suffix))
            shutil.move(self._output_path, os.path.join(directory, self.output_file + self.suffix))
            shutil.move(self._qclog_path, os.path.join(directory, self.qclog_file + self.suffix))
            for file in ("HESS", "GRAD", "dens.0.cube"):
                if os.path.isfile(file):
                    shutil.move(file, file + self.suffix)
        if not self.save_scratch:
            try:
                shutil.rmtree(scratch_dir)
            except FileNotFoundError:
                pass

    def run(self, directory: str | Path = "./"):
        """
        Perform the actual QChem run.

        Args:
            directory (str): The directory to run in. Defaults to "./".

        Returns:
            (subprocess.Popen) Used for monitoring.
        """
        local_scratch = os.path.join(os.environ["QCLOCALSCR"], "scratch")
        if os.path.exists(local_scratch):
            shutil.rmtree(local_scratch)
        if os.path.exists(os.path.join(os.environ["QCSCRATCH"], "132.0")):
            os.mkdir(local_scratch)
            shutil.move(os.path.join(os.environ["QCSCRATCH"], "132.0"), local_scratch)
        if os.path.exists(os.path.join(os.environ["QCSCRATCH"], "53.0")):
            os.makedirs(local_scratch, exist_ok=True)
            shutil.move(os.path.join(os.environ["QCSCRATCH"], "53.0"), local_scratch)
        with open(self.qclog_file, "w") as qc_log:
            return subprocess.Popen(self.current_command, cwd=directory, stdout=qc_log, shell=True)  # pylint: disable=R1732

    @classmethod
    def opt_with_frequency_flattener(
        cls,
        qchem_command,
        multimode="openmp",
        input_file="mol.qin",
        output_file="mol.qout",
        qclog_file="mol.qclog",
        max_iterations=10,
        max_molecule_perturb_scale=0.3,
        check_connectivity=True,
        linked=True,
        transition_state=False,
        freq_before_opt=False,
        save_final_scratch=False,
        **QCJob_kwargs,
    ):
        """
        Optimize a structure and calculate vibrational frequencies to check if the
        structure is in a true minima.

        If there are an inappropriate number of imaginary frequencies (>0 for a
         minimum-energy structure, >1 for a transition-state), attempt to re-calculate
         using one of two methods:
            - Perturb the geometry based on the imaginary frequencies and re-optimize
            - Use the exact Hessian to inform a subsequent optimization
         After each geometry optimization, the frequencies are re-calculated to
         determine if a true minimum (or transition-state) has been found.

        Note: Very small imaginary frequencies (-15cm^-1 < nu < 0) are allowed
        if there is only one more than there should be. In other words, if there
        is one very small imaginary frequency, it is still treated as a minimum,
        and if there is one significant imaginary frequency and one very small
        imaginary frequency, it is still treated as a transition-state.

        Args:
            qchem_command (str): Command to run QChem.
            multimode (str): Parallelization scheme, either openmp or mpi.
            input_file (str): Name of the QChem input file.
            output_file (str): Name of the QChem output file.
            qclog_file (str): Name of the file to redirect the standard output
            max_iterations (int): Number of perturbation -> optimization -> frequency
                iterations to perform. Defaults to 10.
            max_molecule_perturb_scale (float): The maximum scaled perturbation that
                can be applied to the molecule. Defaults to 0.3.
            check_connectivity (bool): Whether to check differences in connectivity
                introduced by structural perturbation. Defaults to True.
            linked (bool): Whether or not to use the linked flattener. If set to True (default),
                then the explicit Hessians from a vibrational frequency analysis will be used
                as the initial Hessian of subsequent optimizations. In many cases, this can
                significantly improve optimization efficiency.
            transition_state (bool): If True (default False), use a ts
                optimization (search for a saddle point instead of a minimum)
            freq_before_opt (bool): If True (default False), run a frequency
                calculation before any opt/ts searches to improve understanding
                of the local potential energy surface.
            save_final_scratch (bool): Whether to save full scratch directory contents
                at the end of the flattening. Defaults to False.
            **QCJob_kwargs: Passthrough kwargs to QCJob. See
                :class:`custodian.qchem.jobs.QCJob`.
        """
        if not os.path.exists(input_file):
            raise AssertionError(f"{input_file=} must be present!")

        if transition_state:
            opt_method = "ts"
            perturb_index = 1
        else:
            opt_method = "opt"
            perturb_index = 0

        energy_diff_cutoff = 0.000001

        orig_input = QCInput.from_file(input_file)
        freq_rem = copy.deepcopy(orig_input.rem)
        freq_rem["job_type"] = "freq"
        opt_rem = copy.deepcopy(orig_input.rem)
        opt_rem["job_type"] = opt_method
        opt_geom_opt = None
        if "geom_opt2" in orig_input.rem:
            freq_rem.pop("geom_opt2", None)
            if linked:
                opt_rem.pop("geom_opt2", None)
        first = True
        energy_history = []

        if freq_before_opt:
            if not linked:
                warnings.warn("WARNING: This first frequency calculation will not inform subsequent optimization!")
            yield (
                QCJob(
                    qchem_command=qchem_command,
                    multimode=multimode,
                    input_file=input_file,
                    output_file=output_file,
                    qclog_file=qclog_file,
                    suffix=".freq_pre",
                    save_scratch=True,
                    backup=first,
                    **QCJob_kwargs,
                )
            )

            freq_outdata = QCOutput(output_file + ".freq_pre").data
            if freq_outdata["version"] == "6":
                opt_set = OptSet(molecule=freq_outdata["initial_molecule"], qchem_version=freq_outdata["version"])
                opt_geom_opt = copy.deepcopy(opt_set.geom_opt)

            if linked:
                opt_rem["geom_opt_hessian"] = "read"
                if freq_outdata["version"] == "6":
                    opt_geom_opt["initial_hessian"] = "read"

            tmp_opt_rem = copy.deepcopy(opt_rem)
            if opt_rem["scf_algorithm"] == "diis":
                tmp_opt_rem["scf_guess_always"] = "True"
            opt_QCInput = QCInput(
                molecule=orig_input.molecule,
                rem=tmp_opt_rem,
                opt=orig_input.opt,
                pcm=orig_input.pcm,
                solvent=orig_input.solvent,
                smx=orig_input.smx,
                vdw_mode=orig_input.vdw_mode,
                van_der_waals=orig_input.van_der_waals,
                nbo=orig_input.nbo,
                geom_opt=opt_geom_opt,
            )
            opt_QCInput.write_file(input_file)
            first = False

        if linked:
            opt_rem["geom_opt_hessian"] = "read"

            for ii in range(max_iterations):
                yield (
                    QCJob(
                        qchem_command=qchem_command,
                        multimode=multimode,
                        input_file=input_file,
                        output_file=output_file,
                        qclog_file=qclog_file,
                        suffix=f".{opt_method}_{ii}",
                        save_scratch=True,
                        backup=first,
                        **QCJob_kwargs,
                    )
                )
                opt_outdata = QCOutput(f"{output_file}.{opt_method}_{ii}").data
                opt_indata = QCInput.from_file(f"{input_file}.{opt_method}_{ii}")
                if opt_outdata["version"] == "6":
                    opt_geom_opt = copy.deepcopy(opt_indata.geom_opt)
                    opt_geom_opt["initial_hessian"] = "read"
                for key in opt_indata.rem:
                    if key not in {"job_type", "geom_opt2", "scf_guess_always"}:
                        if freq_rem.get(key, None) != opt_indata.rem[key] and "geom_opt" not in key:
                            freq_rem[key] = opt_indata.rem[key]
                        if opt_rem.get(key, None) != opt_indata.rem[key]:
                            opt_rem[key] = opt_indata.rem[key]
                first = False
                if (
                    opt_outdata["structure_change"] == "unconnected_fragments"
                    and not opt_outdata["completion"]
                    and not transition_state
                ):
                    warnings.warn(
                        "Unstable molecule broke into unconnected fragments which failed to optimize! Exiting..."
                    )
                    break
                energy_history.append(opt_outdata.get("final_energy"))
                freq_QCInput = QCInput(
                    molecule=opt_outdata.get("molecule_from_optimized_geometry"),
                    rem=freq_rem,
                    opt=orig_input.opt,
                    pcm=orig_input.pcm,
                    solvent=orig_input.solvent,
                    smx=orig_input.smx,
                    vdw_mode=orig_input.vdw_mode,
                    van_der_waals=orig_input.van_der_waals,
                    nbo=orig_input.nbo,
                )
                freq_QCInput.write_file(input_file)
                yield (
                    QCJob(
                        qchem_command=qchem_command,
                        multimode=multimode,
                        input_file=input_file,
                        output_file=output_file,
                        qclog_file=qclog_file,
                        suffix=f".freq_{ii}",
                        save_scratch=True,
                        backup=first,
                        **QCJob_kwargs,
                    )
                )

                freq_outdata = QCOutput(f"{output_file}.freq_{ii}").data
                freq_indata = QCInput.from_file(f"{input_file}.freq_{ii}")
                for key in freq_indata.rem:
                    if key not in {"job_type", "geom_opt2", "scf_guess_always"}:
                        if freq_rem.get(key, None) != freq_indata.rem[key]:
                            freq_rem[key] = freq_indata.rem[key]
                        if opt_rem.get(key, None) != freq_indata.rem[key] and key != "cpscf_nseg":
                            opt_rem[key] = freq_indata.rem[key]
                errors = freq_outdata.get("errors")

                if len(errors) != 0:
                    raise AssertionError("No errors should be encountered while flattening frequencies!")

                if not transition_state:
                    freq_list = freq_outdata.get("frequencies")

                    if len(freq_list) > 1:
                        freq_0 = freq_list[0]
                        freq_1 = freq_list[1]
                    else:
                        freq_0 = freq_outdata.get("frequencies")[0]
                        freq_1 = 100000.0
                        warnings.warn("Only single frequency. Two atom fragment")
                        break

                    if freq_0 > 0.0:
                        warnings.warn("All frequencies positive!")
                        break
                    if abs(freq_0) < 15.0 and freq_1 > 0.0:
                        warnings.warn("One negative frequency smaller than 15.0 - not worth further flattening!")
                        break
                    if len(energy_history) > 1 and abs(energy_history[-1] - energy_history[-2]) < energy_diff_cutoff:
                        warnings.warn("Energy change below cutoff!")
                        break
                    tmp_opt_rem = copy.deepcopy(opt_rem)
                    if opt_rem["scf_algorithm"] == "diis":
                        tmp_opt_rem["scf_guess_always"] = "True"
                    opt_QCInput = QCInput(
                        molecule=opt_outdata.get("molecule_from_optimized_geometry"),
                        rem=tmp_opt_rem,
                        opt=orig_input.opt,
                        pcm=orig_input.pcm,
                        solvent=orig_input.solvent,
                        smx=orig_input.smx,
                        vdw_mode=orig_input.vdw_mode,
                        van_der_waals=orig_input.van_der_waals,
                        nbo=orig_input.nbo,
                        geom_opt=opt_geom_opt,
                    )
                    opt_QCInput.write_file(input_file)
                else:
                    freq_0 = freq_outdata.get("frequencies")[0]
                    freq_1 = freq_outdata.get("frequencies")[1]
                    freq_2 = freq_outdata.get("frequencies")[2]
                    if freq_0 < 0.0 < freq_1:
                        warnings.warn("Saddle point found!")
                        break
                    if abs(freq_1) < 15.0 and freq_2 > 0.0:
                        warnings.warn(
                            "Second small imaginary frequency (smaller than 15.0) - not worth further flattening!"
                        )
                        break
                    tmp_opt_rem = copy.deepcopy(opt_rem)
                    if opt_rem["scf_algorithm"] == "diis":
                        tmp_opt_rem["scf_guess_always"] = "True"
                    opt_QCInput = QCInput(
                        molecule=opt_outdata.get("molecule_from_optimized_geometry"),
                        rem=tmp_opt_rem,
                        opt=orig_input.opt,
                        pcm=orig_input.pcm,
                        solvent=orig_input.solvent,
                        smx=orig_input.smx,
                        vdw_mode=orig_input.vdw_mode,
                        van_der_waals=orig_input.van_der_waals,
                        nbo=orig_input.nbo,
                        # geom_opt=opt_geom_opt, # Will be uncommented once new optimizer supports TS calcs
                    )
                    opt_QCInput.write_file(input_file)
            if not save_final_scratch:
                shutil.rmtree(os.path.join(os.getcwd(), "scratch"))

        else:
            orig_opt_input = QCInput.from_file(input_file)
            history = []

            for ii in range(max_iterations):
                yield (
                    QCJob(
                        qchem_command=qchem_command,
                        multimode=multimode,
                        input_file=input_file,
                        output_file=output_file,
                        qclog_file=qclog_file,
                        suffix=f".{opt_method}_{ii}",
                        backup=first,
                        **QCJob_kwargs,
                    )
                )
                opt_outdata = QCOutput(f"{output_file}.{opt_method}_{ii}").data
                if first:
                    orig_species = copy.deepcopy(opt_outdata.get("species"))
                    orig_charge = copy.deepcopy(opt_outdata.get("charge"))
                    orig_multiplicity = copy.deepcopy(opt_outdata.get("multiplicity"))
                    orig_energy = copy.deepcopy(opt_outdata.get("final_energy"))
                first = False
                if (
                    opt_outdata["structure_change"] == "unconnected_fragments"
                    and not opt_outdata["completion"]
                    and not transition_state
                ):
                    warnings.warn(
                        "Unstable molecule broke into unconnected fragments which failed to optimize! Exiting..."
                    )
                    break
                freq_QCInput = QCInput(
                    molecule=opt_outdata.get("molecule_from_optimized_geometry"),
                    rem=freq_rem,
                    opt=orig_opt_input.opt,
                    pcm=orig_opt_input.pcm,
                    solvent=orig_opt_input.solvent,
                    smx=orig_opt_input.smx,
                    vdw_mode=orig_opt_input.vdw_mode,
                    van_der_waals=orig_opt_input.van_der_waals,
                    nbo=orig_input.nbo,
                )
                freq_QCInput.write_file(input_file)
                yield (
                    QCJob(
                        qchem_command=qchem_command,
                        multimode=multimode,
                        input_file=input_file,
                        output_file=output_file,
                        qclog_file=qclog_file,
                        suffix=f".freq_{ii}",
                        backup=first,
                        **QCJob_kwargs,
                    )
                )
                outdata = QCOutput(f"{output_file}.freq_{ii}").data
                indata = QCInput.from_file(f"{input_file}.freq_{ii}")
                if "cpscf_nseg" in indata.rem:
                    freq_rem["cpscf_nseg"] = indata.rem["cpscf_nseg"]
                errors = outdata.get("errors")
                if len(errors) != 0:
                    raise AssertionError("No errors should be encountered while flattening frequencies!")
                if not transition_state:
                    freq_0 = outdata.get("frequencies")[0]
                    freq_1 = outdata.get("frequencies")[1]
                    if freq_0 > 0.0:
                        warnings.warn("All frequencies positive!")
                        if opt_outdata.get("final_energy") > orig_energy:  # type: ignore
                            warnings.warn("WARNING: Energy increased during frequency flattening!")
                        break
                    if abs(freq_0) < 15.0 and freq_1 > 0.0:
                        warnings.warn("One negative frequency smaller than 15.0 - not worth further flattening!")
                        break
                    if len(energy_history) > 1 and abs(energy_history[-1] - energy_history[-2]) < energy_diff_cutoff:
                        warnings.warn("Energy change below cutoff!")
                        break
                else:
                    freq_0 = outdata.get("frequencies")[0]
                    freq_1 = outdata.get("frequencies")[1]
                    freq_2 = outdata.get("frequencies")[2]
                    if freq_0 < 0.0 < freq_1:
                        warnings.warn("Saddle point found!")
                        break
                    if abs(freq_1) < 15.0 and freq_2 > 0.0:
                        warnings.warn(
                            "Second small imaginary frequency (smaller than 15.0) - not worth further flattening!"
                        )
                        break

                hist = {}
                hist["molecule"] = copy.deepcopy(outdata.get("initial_molecule"))
                hist["geometry"] = copy.deepcopy(outdata.get("initial_geometry"))
                hist["frequencies"] = copy.deepcopy(outdata.get("frequencies"))
                hist["frequency_mode_vectors"] = copy.deepcopy(outdata.get("frequency_mode_vectors"))
                hist["num_neg_freqs"] = sum(1 for freq in outdata.get("frequencies") if freq < 0)
                hist["energy"] = copy.deepcopy(opt_outdata.get("final_energy"))
                hist["index"] = len(history)
                hist["children"] = []
                history.append(hist)

                ref_mol = history[-1]["molecule"]
                geom_to_perturb = history[-1]["geometry"]
                negative_freq_vecs = history[-1]["frequency_mode_vectors"][perturb_index]
                reversed_direction = False
                standard = True

                # If we've found one or more negative frequencies in two consecutive iterations, let's dig in
                # deeper:
                if len(history) > 1:
                    # Start by finding the latest iteration's parent:
                    if history[-1]["index"] in history[-2]["children"]:
                        parent_hist = history[-2]
                        history[-1]["parent"] = parent_hist["index"]
                    elif history[-1]["index"] in history[-3]["children"]:
                        parent_hist = history[-3]
                        history[-1]["parent"] = parent_hist["index"]
                    else:
                        raise AssertionError(
                            "ERROR: your parent should always be one or two iterations behind you! Exiting..."
                        )

                    # if the number of negative frequencies has remained constant or increased from parent to
                    # child,
                    if history[-1]["num_neg_freqs"] >= parent_hist["num_neg_freqs"]:
                        # check to see if the parent only has one child, aka only the positive perturbation has
                        # been tried,
                        # in which case just try the negative perturbation from the same parent
                        if len(parent_hist["children"]) == 1:
                            ref_mol = parent_hist["molecule"]
                            geom_to_perturb = parent_hist["geometry"]
                            negative_freq_vecs = parent_hist["frequency_mode_vectors"][perturb_index]
                            reversed_direction = True
                            standard = False
                            parent_hist["children"].append(len(history))
                        # If the parent has two children, aka both directions have been tried, then we have to
                        # get creative:
                        elif len(parent_hist["children"]) == 2:
                            # If we're dealing with just one negative frequency,
                            if parent_hist["num_neg_freqs"] == 1:
                                if history[parent_hist["children"][0]]["energy"] < history[-1]["energy"]:
                                    good_child = copy.deepcopy(history[parent_hist["children"][0]])
                                else:
                                    good_child = copy.deepcopy(history[-1])
                                if good_child["num_neg_freqs"] > 1:
                                    raise ValueError(
                                        "ERROR: Child with lower energy has more negative frequencies! Exiting..."
                                    )
                                if good_child["energy"] < parent_hist["energy"] or (
                                    vector_list_diff(
                                        good_child["frequency_mode_vectors"][perturb_index],
                                        parent_hist["frequency_mode_vectors"][perturb_index],
                                    )
                                    > 0.2
                                ):
                                    make_good_child_next_parent = True
                                else:
                                    raise ValueError("ERROR: Good child not good enough! Exiting...")
                                if make_good_child_next_parent:
                                    good_child["index"] = len(history)
                                    history.append(good_child)
                                    ref_mol = history[-1]["molecule"]
                                    geom_to_perturb = history[-1]["geometry"]
                                    negative_freq_vecs = history[-1]["frequency_mode_vectors"][perturb_index]
                            else:
                                raise ValueError("ERROR: Can't deal with multiple neg frequencies yet! Exiting...")
                        else:
                            raise ValueError("ERROR: Parent cannot have more than two children! Exiting...")
                    # Implicitly, if the number of negative frequencies decreased from parent to child,
                    # continue normally.
                if standard:
                    history[-1]["children"].append(len(history))

                min_molecule_perturb_scale = 0.1
                scale_grid = 10
                perturb_scale_grid = (max_molecule_perturb_scale - min_molecule_perturb_scale) / scale_grid

                structure_successfully_perturbed = False
                for molecule_perturb_scale in np.arange(
                    max_molecule_perturb_scale,
                    min_molecule_perturb_scale,
                    -perturb_scale_grid,
                ):
                    new_coords = perturb_coordinates(
                        old_coords=geom_to_perturb,
                        negative_freq_vecs=negative_freq_vecs,
                        molecule_perturb_scale=molecule_perturb_scale,
                        reversed_direction=reversed_direction,
                    )
                    new_molecule = Molecule(
                        species=orig_species,  # type: ignore
                        coords=new_coords,
                        charge=orig_charge,  # type: ignore
                        spin_multiplicity=orig_multiplicity,  # type: ignore
                    )
                    if check_connectivity and not transition_state:
                        structure_successfully_perturbed = (
                            check_for_structure_changes(ref_mol, new_molecule) == "no_change"
                        )
                        if structure_successfully_perturbed:
                            break
                if not structure_successfully_perturbed:
                    raise Exception(
                        "ERROR: Unable to perturb coordinates to remove negative frequency without changing "
                        "the connectivity! Exiting..."
                    )

                new_opt_QCInput = QCInput(
                    molecule=new_molecule,  # type: ignore
                    rem=opt_rem,
                    opt=orig_opt_input.opt,
                    pcm=orig_opt_input.pcm,
                    solvent=orig_opt_input.solvent,
                    smx=orig_opt_input.smx,
                    vdw_mode=orig_opt_input.vdw_mode,
                    van_der_waals=orig_opt_input.van_der_waals,
                    nbo=orig_input.nbo,
                    geom_opt=orig_input.geom_opt,
                )
                new_opt_QCInput.write_file(input_file)
